#include "config.h"
#include "utils.h"
#include "neighbour.h"
#include "particledata.h"
#include "tree.h"

#include <cmath>
#include <vector>
#include <iostream>
#include <algorithm>

using std::vector; 
using std::max; 

typedef vector<ParticleData>::size_type p_sz; 
typedef vector<ParticleList>::size_type pl_sz; 
typedef vector<NodeInfo>::size_type nd_sz; 
//----------------------------------------------------------------------
bool isOverlap(const NodeInfo& node, const real *x, const double h); 
bool isOverlapPair(const NodeInfo& node, const real *pos, const double h); 
//----------------------------------------------------------------------

void FindNeighbours(vector<p_sz>& nlist, const vector<NodeInfo>& tree, const vector<ParticleList>& plist, 
    const vector<ParticleData>& sph, const real *pos, const double h)
{
  nd_sz root = 0; 
  nd_sz current_node_index = tree[root].GetChild(); 

  double h2  = h * h; 
  nlist.clear(); // initialize the neighbour list. 

  while (current_node_index != root) {
    const NodeInfo& node = tree[current_node_index]; 
    if ( !isOverlap(node, pos, h) ) {
      current_node_index = node.GetNext(); 
    } else if (node.GetChild() != NONE) {
      current_node_index = node.GetChild(); 
    } else {
      for (pl_sz p = node.GetTopLeaf(); p != node.GetTopLeaf() + node.GetNumberOfLeaves(); ++p) {
        p_sz j = plist[p].GetIndex(); 
        double dl2 = 0; 
#ifndef PERIODIC
        for (int dim = 0; dim != DIM; ++dim) 
          dl2 += sq(sph[j].GetPos(dim) - pos[dim]); 
#else 
        double dx[DIM]; 
        for (int dim = 0; dim != DIM; ++dim) {
          dx[dim] = sph[j].GetPos(dim) - pos[dim]; 
          periodic_dx(dx[dim]); 
          dl2 += sq(dx[dim]); 
        }
#endif
        if (dl2 <= h2) 
          nlist.push_back(j); 
      }
      current_node_index = node.GetNext(); 
    }
  }
}

void FindNeighbourPairs(vector<p_sz>& nlist, const vector<NodeInfo>& tree, const vector<ParticleList>& plist, 
    const vector<ParticleData>& sph, const real *pos, const double h)
{
  nd_sz root = 0; 
  nd_sz current_node_index = tree[root].GetChild(); 

  double h2  = h * h; 
  nlist.clear(); // initialize the neighbour list. 

  while (current_node_index != root) {
    const NodeInfo& node = tree[current_node_index]; 
    if ( !isOverlapPair(node, pos, h) ) {
      current_node_index = node.GetNext(); 
    } else if (node.GetChild() != NONE) {
      current_node_index = node.GetChild(); 
    } else {
      for (pl_sz p = node.GetTopLeaf(); p != node.GetTopLeaf() + node.GetNumberOfLeaves(); ++p) {
        p_sz j = plist[p].GetIndex(); 
        double dl2 = 0; 
#ifndef PERIODIC
        for (int dim = 0; dim != DIM; ++dim) 
          dl2 += sq(sph[j].GetPos(dim) - pos[dim]); 
#else 
        double dx[DIM]; 
        for (int dim = 0; dim != DIM; ++dim) {
          dx[dim] = sph[j].GetPos(dim) - pos[dim]; 
          periodic_dx(dx[dim]); 
          dl2 += sq(dx[dim]); 
        }
#endif
        if (dl2 <= h2 || dl2 < sq(sph[j].GetHsml())); 
          nlist.push_back(j); 
      }
      current_node_index = node.GetNext(); 
    }
  }
}

bool isOverlap(const NodeInfo& node, const real *pos, const double h)
{
  const double hwidth = 5.e-1 * node.GetWidth(); 
  double r2 = 0; 
  for (int dim = 0; dim != DIM; ++dim) {
    double dl = pos[dim] - node.GetPos(dim); 
#ifdef PERIODIC
    periodic_dx(dl); 
#endif 
    dl = std::max(0.0, fabs(dl) - hwidth); 
    r2 += sq(dl); 
  }

  if (r2 > h *h ) 
    return false; 
  else 
    return true; 
}

bool isOverlapPair(const NodeInfo& node, const real *pos, const double h)
{
  const double hwidth = 5.e-1 * node.GetWidth(); 
  double r2 = 0; 
  for (int dim = 0; dim != DIM; ++dim) {
    double dl = pos[dim] - node.GetPos(dim); 
#ifdef PERIODIC
    periodic_dx(dl); 
#endif 
    dl = std::max(0.0, fabs(dl) - hwidth); 
    r2 += sq(dl); 
  }

  double hmax = max(h, node.GetHsml()); // node.GetHsml() returns the maximum smoothing length in the node 

  if (r2 > sq(hmax)) 
    return false; 
  else 
    return true; 
}
