#include "config.h"
#include "particledata.h"
#include "tree.h"

#include <iostream>
#include <cmath>
#include <vector>
#include <algorithm>

using std::vector; 
using std::max; 
using std::min; 
using std::cout; 
using std::cin; 
using std::endl; 

typedef vector<ParticleList>::size_type pl_sz; 
typedef vector<NodeInfo>::size_type nd_sz;
typedef vector<ParticleData>::size_type p_sz; 

void make_particle_list(vector<ParticleList>& plist, const vector<ParticleData>& sph); 
void make_root(vector<NodeInfo>& tree, vector<ParticleList>& plist, const vector<ParticleData>& sph); 
void construct_tree(vector<NodeInfo>& tree, vector<ParticleList>& plist, const vector<ParticleData>& sph); 
bool terminate_tree(const NodeInfo& node); 
void make_next(vector<NodeInfo>& tree); 
void deallocate_sublist(pl_sz *sublist[] ); 
nd_sz find_next(nd_sz current_node_index, const vector<NodeInfo>& tree); 
inline int bitsign(const int i); 
void calc_hmax(vector<NodeInfo>& tree, const vector<ParticleList>& plist, 
    const vector<ParticleData>& sph, const nd_sz current_node_index); 

ParticleList::ParticleList() 
{
  index = 0; 
}


NodeInfo::NodeInfo () 
{
  next = parent = child = brother = NONE; 
  number = 0; 
  top_leaf = NONE; 
  max_hsml = 0; 
}

void BuildTree(vector<NodeInfo>& tree, vector<ParticleList>& plist, const vector<ParticleData>& sph) 
{
  if (plist.capacity() < sph.size()) 
    plist.resize(sph.size()); 

  // initialize
  plist.clear(); 
  tree.clear(); 

  make_particle_list(plist, sph); 
  make_root(tree, plist, sph); 
  construct_tree(tree, plist, sph); 
  make_next(tree); 
}

void make_particle_list(vector<ParticleList>& plist, const vector<ParticleData>& sph) 
{
  for (p_sz i = 0; i != sph.size(); ++i) {
    ParticleList tmp; 
    tmp.PutIndex(i); 
    plist.push_back(tmp); 
  }
}

void make_root(vector<NodeInfo>& tree, vector<ParticleList>& plist, const vector<ParticleData>& sph)
{
  real min_pos[DIM], max_pos[DIM]; 
#ifdef PERIODIC
  for (int dim = 0; dim != DIM; ++dim) {
    min_pos[dim] = 0; 
    max_pos[dim] = BoxSize; 
  }
#endif 

#ifndef PERIODIC
  for (int dim = 0; dim != DIM; ++dim) 
    min_pos[dim] = max_pos[dim] = sph[0].GetPos(dim); 
 
  
  for (p_sz i = 1; i != sph.size(); ++i) {
    for (int dim = 0; dim != DIM; ++dim)  {
      min_pos[dim] = min(min_pos[dim], sph[i].GetPos(dim)); 
      max_pos[dim] = max(max_pos[dim], sph[i].GetPos(dim)); 
    }
  }
#endif 

  NodeInfo node; 

  // define width
#ifdef PERIODIC
  node.PutWidth(static_cast<double>(BoxSize)); 
#else 
  node.PutWidth(0); 
  for (int dim = 0; dim != DIM; ++dim) 
    node.PutWidth(max(node.GetWidth(), max_pos[dim] - min_pos[dim])); 
#endif

  for (int dim = 0; dim != DIM; ++dim) 
    node.PutPos(5.e-1 * (max_pos[dim] + min_pos[dim]), dim);  

  node.PutTopLeaf(0); 
  node.PutNumberOfLeaves(plist.size()); 
  node.PutHsml(0); 
  for (p_sz i = 0; i != sph.size(); ++i) 
    node.PutHsml(max(node.GetHsml(), sph[i].GetHsml())); 

  node.PutLevel(0); 

  tree.push_back(node); 
}



void construct_tree(vector<NodeInfo>& tree, vector<ParticleList>& plist, const vector<ParticleData>& sph) 
{

  static pl_sz *sublist[nsub]; 
  pl_sz n_part_in_sub[nsub]; 
  for (int i = 0; i != nsub; ++i)
    sublist[i] = new pl_sz[plist.size()]; 


  const nd_sz root = 0; 
  nd_sz current_node_index = root; // start from root 


  while (1) { // endless loop 

    if (terminate_tree(tree[current_node_index]) && current_node_index != root ) { 
      // Now we stop making children
      if (tree[current_node_index].GetBrother() != NONE) {
        current_node_index = tree[current_node_index].GetBrother(); 
      } else { // nothing to do 
        while (1) {
          if (tree[tree[current_node_index].GetParent()].GetBrother() != NONE) { 
            // go to the younger brother of the parent
            current_node_index =  tree[tree[current_node_index].GetParent()].GetBrother(); 
            break; 
          } else if (tree[current_node_index].GetParent() == root) {
            // ok. the parent is ROOT. Get out of here!
            deallocate_sublist(sublist); 
            return; 
          }
          current_node_index = tree[current_node_index].GetParent(); 
        }
      }
      continue; 
    }

    for (int sub = 0; sub != nsub; ++sub) 
      n_part_in_sub[sub] = 0; 


    for (pl_sz leaf = tree[current_node_index].GetTopLeaf(); 
        leaf != tree[current_node_index].GetTopLeaf() + tree[current_node_index].GetNumberOfLeaves(); ++leaf) {
      p_sz particle_index = plist[leaf].GetIndex(); 
      int subindex = 0; 
      for (int dim = 0; dim != DIM; ++dim) { 
        if (tree[current_node_index].GetPos(dim) <= sph[particle_index].GetPos(dim)) {
          subindex += 1 << dim; 
        }
      }

      sublist[subindex][n_part_in_sub[subindex]] = particle_index; 
      n_part_in_sub[subindex] ++; 
    }

    nd_sz next_node_index;  
    for (int sub = 0; sub != nsub; ++sub) {
      if (n_part_in_sub[sub] != 0) { // this subnode contains particles 
        NodeInfo child_node; 
        child_node.PutParent(current_node_index); 
        if (tree[current_node_index].GetChild() == NONE) { 
          // current node does not have any children so far 
          tree[current_node_index].PutChild(tree.size()); 
          child_node.PutTopLeaf(tree[current_node_index].GetTopLeaf()); 
          next_node_index = tree[current_node_index].GetChild(); 
        } else {
          // this one should have an older brother 
          // and the older brother lives in the tail of the tree vector
          NodeInfo& senior = tree[tree.size() - 1]; 
          child_node.PutTopLeaf(senior.GetTopLeaf() + senior.GetNumberOfLeaves()); 
          senior.PutBrother(tree.size()); 
        } 
        const NodeInfo& father = tree[current_node_index]; 
        child_node.PutNumberOfLeaves(n_part_in_sub[sub]); 
        child_node.PutWidth(5.e-1 * father.GetWidth()); 
        child_node.PutLevel(father.GetLevel() + 1); 

        // rearrange leaves 
        for (pl_sz leaf = 0; leaf != n_part_in_sub[sub]; ++leaf) 
          plist[child_node.GetTopLeaf() + leaf].PutIndex(sublist[sub][leaf]); 

        // define the center of the child node
        for (int dim = 0; dim != DIM; ++dim) 
          child_node.Pos(dim) = tree[current_node_index].GetPos(dim) 
            + bitsign( (sub>>dim) & 1) * 2.5e-1 * tree[current_node_index].GetWidth(); 

        tree.push_back(child_node); 
      }
    }
    current_node_index = next_node_index; 
  } 
}


void deallocate_sublist(pl_sz *sublist[] ) 
{
  for (int i = 0; i != nsub; ++i) {
    int ii = nsub - 1 - i; 
    delete [] sublist[ii]; 
  }
}

void make_next(vector<NodeInfo>& tree) 
{
  for (nd_sz node = 1; node != tree.size(); ++node)
    tree[node].PutNext(find_next(node, tree)); 
}

nd_sz find_next(nd_sz current_node_index, const vector<NodeInfo>& tree)
{
  nd_sz root = 0; 

  const nd_sz nd_backup = current_node_index; 

  if(tree[current_node_index].GetBrother() != NONE) {
    // this node has a brother
    current_node_index = tree[current_node_index].GetBrother(); 
  } else {
    nd_sz next_node_index = current_node_index; 
    while(1) {
      if (tree[tree[next_node_index].GetParent()].GetBrother() != NONE) {
        // the younger brother of the paranet must be the next!
        current_node_index = tree[tree[next_node_index].GetParent()].GetBrother(); 
        break; 
      } else if (tree[next_node_index].GetParent() == root) {
        current_node_index = root; 
        break; 
      }
      next_node_index = tree[next_node_index].GetParent(); 
    }
  }
  if (current_node_index == nd_backup) 
    cout << "Node: " << nd_backup << "'s Next is " << current_node_index << endl; 
  return current_node_index; 
}

void UpdateHmax(vector<NodeInfo>& tree, const vector<ParticleList>& plist, const vector<ParticleData>& sph)
{
  for (nd_sz node = 0; node != tree.size(); ++node) 
    tree[node].PutHsml(0); 
  calc_hmax(tree, plist, sph, 0); 
}

// find the maximum smoothing length of the particles belonging to the node recursively. 
void calc_hmax(vector<NodeInfo>& tree, const vector<ParticleList>& plist, 
    const vector<ParticleData>& sph, const nd_sz current_node_index) 
{
  
  
  NodeInfo& now = tree[current_node_index]; 
  nd_sz child_node_index = tree[current_node_index].GetChild(); 

  if (child_node_index == NONE) {
    // this node is placed at the bottom of the tree
    double hmax = sph[now.GetTopLeaf()].GetHsml(); 
    for (pl_sz leaf = now.GetTopLeaf()+1; leaf != now.GetTopLeaf() + now.GetNumberOfLeaves(); ++leaf) {
      hmax = max(hmax, sph[leaf].GetHsml()); 
    }
    now.PutHsml(hmax); 
    if (now.GetBrother() != NONE) {
      calc_hmax(tree, plist, sph, now.GetBrother()); 
    } else if (now.GetParent() != 0) { // its parent is not the ROOT
      calc_hmax(tree, plist, sph, now.GetParent()); 
    } else {
      return; 
    }
  } else if (tree[child_node_index].GetHsml() == 0) {
    calc_hmax(tree, plist, sph, child_node_index); 
  } else {
    double hmax = tree[child_node_index].GetHsml(); 
    while (tree[child_node_index].GetBrother() != NONE) {
      child_node_index = tree[child_node_index].GetBrother(); 
      hmax = max(hmax, tree[child_node_index].GetHsml()); 
    }
    now.PutHsml(hmax); 

    if (now.GetBrother() != NONE) {
      calc_hmax(tree, plist, sph, now.GetBrother()); 
    } else if (now.GetParent() != 0) {
      calc_hmax(tree, plist, sph, now.GetParent()); 
    } else {
      return; 
    }
  }
}

bool terminate_tree(const NodeInfo& node) 
{
  if (node.GetNumberOfLeaves() <= NodeCritNum || node.GetLevel() > MaxNodeLevel)
    return true; 
  else 
    return false; 
}

inline int bitsign(const int i) 
{
    return ( (i == 1) ? (1) : (-1) ); 
}

